#ifndef CPPTAGDB_HEADER_FILE_INCLUDED_P
#define CPPTAGDB_HEADER_FILE_INCLUDED_P
//
// Copyright 2011, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

///@file  cxxtls/cpptagdb.h  Symbols needed to access the output of cpptagdb.exe

#include <vector>
#include <map>
#include <set>
#include <iostream>

#include <string>

namespace cxxtls
{
namespace CppTagDB
{

    struct SymbolInfo
       ///  A description of a symbol as extracted by the cpptagdb.exe program.
       ///  The description includes the type of the symbol, its full namespace
       ///  name, its member name, the file and line where it is defined, and
       ///  the kind of symbol it is:  function, variable, namespace, etc.
    {
       enum Type
       {
           Function,
           Variable,
           Class,       // or struct or union
           Enum,        // name of an enumeration type
           Enumeration, // enumeration value
           Typedef,
           Namespace,
           FileScope,    // a file containing scopes
           Used,         // using symbol;
           Define,       // #define name
           unknownType,
       };

       std::string fullname_;   ///<  full symbol name:  @b ::namespace::class::member
       std::string member_;     ///<  member name only:  @b member
       std::string type_;       ///<  cpptagdb.exe supplied type string
       std::string file_;       ///<  name of the file defining the symbol
       int         line_;       ///<  line within file_ where the symbol is defined
       Type        typeId_;     ///<  integer "type" of the symbol

       SymbolInfo()
       : line_(0)
       {
       }

       SymbolInfo(std::string const &symbol, 
                  std::string const &member, 
                  std::string const &type,
                  std::string const &file,
                  int                line
                 )
       : fullname_(symbol)
       , member_(member)
       , type_(type)
       , file_(file)
       , line_(line)
       , typeId_(computeTypeId(type))
       {
       }

       typedef std::multimap<std::string, SymbolInfo> Children; // keyed on member name

       Children children_;  ///< nested direct submembers of this symbol, assuming it is a
                            ///  class, struct, union, or namespace

       typedef Children::iterator iterator;             // just for convenience
       typedef Children::const_iterator const_iterator; // just for convenience

       void output(std::ostream &s, std::string indent) const;
           ///<  Output this SymbolInfo object to a stream and
           ///   include all its children indented underneath it.
           ///   A start indentation string should be provided.

       friend std::ostream &operator<< (std::ostream &s, SymbolInfo const &me)
          ///<  Global free functioon to print SymbolInfo objects to
          ///   ostreams.
          ///
          ///   @return the specified stream
          ///   @param s  The ostream stream on which data will be written
          ///   @param me The SymbolInfo object to write to the stream.
          ///   
       {
          me.output(s, "");
          return s;
       }


       Type computeTypeId(std::string const &typeName);
           ///<  Update the typeId_ member appropriately based on the name of
           ///   the symbol type identifier read from the .tagpp file.

       char typeChar() const;  ///<  return a 1 character type id for this member

       std::string annotatedName(bool member=false) const; ///<  return the member name with annotations


    };

class Info
{
public:
    typedef std::multimap<std::string, std::string> FilesMap;  ///< data structure to hold the
                                                               ///  names of the files processed
                                                               ///  by cpptagdb.exe keyed by 
                                                               ///  their basename.ext component.

    typedef std::multimap<std::string, SymbolInfo const *> MembersMap;  ///< a mapping between
                                                                        ///  class and namespace
                                                                        ///  members and their
                                                                        ///  symbol info.

    typedef std::set<std::string> Directories; ///<  the names of all directories encountered
                                               ///   when populating symbols into
                                               ///   members_

    SymbolInfo symbols_;  ///< A tree structure containing all the symbols
                          ///  read by cpptagdb.exe


    FilesMap files_;  ///<  The files processed by cpptagdb.exe. Note that the key is the file node
                      ///   name (basename.ext).  

    Directories directories_; ///<  All the files encountered while processing symbols.

    MembersMap members_; ///< namespace and class members 

    void findMatchingFiles(std::string const &fileNodeName, std::vector<std::string> &output) const;
        ///<  Search the set of files found by cpptagdb.exe and append the name of each
        ///   file whose basename.ext matches fileNodeName to a vector.  Wildcards are @b not
        ///   supported.  fileNodeName must not have directory components.
        ///
        ///  @param[in]  fileNodeName  The name of the file to search for.
        ///  @param[out] output        The vector to which the matching files are @b appended.
        ///

    void findMatchingMembers(std::string const &memberName, std::vector<SymbolInfo const *> &output, bool caseSensitive=true) const;
        ///<  Search this Info object for the names of symbols whose member name
        ///   matches the specified value and append their names to the output
        ///   vector.
        ///
        ///  @param[in]  memberName    The name of the member to search for.
        ///  @param[out] output        The vector to which the matching symbols are @b appended.
        ///

    void findRelativeFiles(std::string const &relativeSubPath, std::vector<std::string> &output) const;
       ///<  Find any files in the directories where symbols are defined which match the specified
       ///   sub path.  For example, if relativeSubPath is fred.cpp, then all fred.cpp files in any
       ///   directory which defined a symbol will be returned.  The relativeSubPath can be an
       ///   absolute path or it can contain ./ or ../ and the function will still work.
       ///   Note that the current directory is explicitly NOT searched by this function.
       ///



    Info() 
    {
    }

    int populate(std::string const &filename, int &failedLine);
       ///< Populate this Info object given a file name.  Return a negative
       /// number if parsing of the file failed and also return the 
       /// line number through the reference parameter, failedLine
       ///


    int populate(FILE *f, int &failedLine); ///<  Read the specified file and populate
                                            ///   this object from that file.
                                            ///   @return negative if read or syntax error.
                                            ///   @return The failing line number in
                                            ///   variable, failedLine.


    friend std::ostream &operator<< (std::ostream &s, Info const &me); ///< global free function
                                                                       ///  that writes Info objects
                                                                       ///  a stream.

private:

    int populateSymbol(std::string const &filename, std::string const &info); ///< Add a symbol to this
                                                                              ///  database given the
                                                                              ///  filename with which it
                                                                              ///  is associated and the
                                                                              ///  line of text from the
                                                                              ///  .tagpp file.
};

} // namespace CppTagDB
} // namespace cxxtls



#endif
