//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/exception.h>
#include <cxxtls/thread.h>
#include <iostream>
#include <errno.h>

#ifndef _MSC_VER
#    include <unistd.h>
#else
#    include <io.h>
#endif



namespace cxxtls
{

    DECLARE_CXXTLS_EXCEPTION(ThreadStaticInitUnfinished);
    
    Mutex coutLock;  // used by CXXTLS_LOG()

    ThreadTraceLevel coutLevel = TT_LOG;

    static int ThreadInitProofFunc();

    static int ThreadInitProof = ThreadInitProofFunc();        // if not yet called, then we have a problem

    static int ThreadInitProofFunc() { return 0x12345897; }

    inline void CHECK_STATIC_INIT()
    {
        if(ThreadInitProof != ThreadInitProofFunc())
        {
           throw ThreadStaticInitUnfinished();
        }
    }


#ifndef _MSC_VER
    ThreadId  createThread(ThreadFunction f, void *d, bool joinable)
    {
        ThreadId id;

        pthread_attr_t attr;

        pthread_attr_init(&attr);

        pthread_attr_setdetachstate(&attr, 
                                    joinable ? (PTHREAD_CREATE_JOINABLE)
                                             : (PTHREAD_CREATE_DETACHED)
                                   );


        if(0 != pthread_create (&id, &attr, f, d))
        {
            CXXTLS_HEAVYDEBUG("Thread Creation failure");
            return 0;

        }


       CXXTLS_HEAVYDEBUG("Thread created: " << id);
       return id;
    }

#else

    ThreadId  createThread(ThreadFunction f, void *d)
    {
        // all threads are joinable in windows.


        HANDLE id;

        if(0 == (id = CreateThread(0, 0, f, d, 0, 0)))
        {
            CXXTLS_HEAVYDEBUG("Thread creation failed");

            return 0;
        }

       CXXTLS_HEAVYDEBUG("Thread created: " << id);
       return id;
    }

#endif

#ifndef _MSC_VER

    bool joinThread(ThreadId id)
    {
        if(id == 0)
        {
            CXXTLS_HEAVYDEBUG("joinThread failure:  " << id);

            return false;  // failure
        }

        if(0 != pthread_join(id, 0))
        {
           CXXTLS_HEAVYDEBUG("joinThread failure:  " << id);
           return false;  // failure

        }

        CXXTLS_HEAVYDEBUG("Thread joined successfully:  " << id);

        return true; // success
    }

#else

    bool joinThread(ThreadId id)
    {

        if(id == 0)
        {
            CXXTLS_HEAVYDEBUG("joinThread Invalid thread failure.");
            return false;  // failure
        }

        DWORD rv = WaitForSingleObject(id, INFINITE);


        if(0 != rv)
        {
           CXXTLS_HEAVYDEBUG("joinThread failure:  id=" << id);
           CXXTLS_HEAVYDEBUG("joinThread failure:  error=" << rv);
           return false; // failure
        }

        CXXTLS_HEAVYDEBUG("Thread joined successfully:  " << id);
        return true;   // success
    }

#endif


   Mutex::Mutex()
   {

#      ifndef _MSC_VER
           if( pthread_mutex_init(&mutex_, 0) )
           {
               throw InitFailureException();
           }
#      else
           if (!InitializeCriticalSectionAndSpinCount(&mutex_, 0x00000400) ) 
           {
               throw InitFailureException();
           }

       
#      endif


   }

   Mutex::~Mutex()
   {

#     ifndef _MSC_VER
          if( pthread_mutex_destroy(&mutex_) )
          {
             // throw DestroyFailureException();  // no can do big guy, Stroustrup explains that
                                                  // destructors are not supposed to throw
                                                  // because it causes memory leaks.
          }
#     else
          if(TryEnterCriticalSection(&mutex_))
          {
             // we either already owned the critical section
             // or got the lock.  In both cases, we should leave it before deleting the mutex
             LeaveCriticalSection(&mutex_);
          }
          DeleteCriticalSection(&mutex_);
#     endif
   }

   void Mutex::lock()
   {
       CHECK_STATIC_INIT();

#      ifndef _MSC_VER
           if( pthread_mutex_lock(&mutex_) )
           {
               throw Mutex::LockFailureException();
           }
#      else
           EnterCriticalSection(&mutex_);
#      endif


   }

   
   bool Mutex::tryLock()
   {
       CHECK_STATIC_INIT();

#      ifndef _MSC_VER
           if( int err = pthread_mutex_trylock(&mutex_) )
           {
              if(err != EBUSY)
              {
                 throw Mutex::LockFailureException();
              }

              return false; // fail:  unable to lock.  Some kind of error, or simply busy 
           }
           else
           {
              return true; // success:  got the lock
           }
#      else
           if(TryEnterCriticalSection(&mutex_))
           {
               return true; // success, got the lock
           }
           else
           {
              return false; // fail, no can do big guy
           }
#      endif


   }

   
   void Mutex::unlock()
   {
       CHECK_STATIC_INIT();

#      ifndef _MSC_VER
           if(pthread_mutex_unlock(&mutex_))
           {
               throw Mutex::UnlockFailureException();
           }
#      else
           LeaveCriticalSection(&mutex_);
#      endif
   }


    ConditionVariable::ConditionVariable()
    {
#       ifndef _MSC_VER
            pthread_cond_init(&condition_, NULL);
#       else
            InitializeConditionVariable(&condition_);
#       endif
    }

    void ConditionVariable::wait(ScopedMutexLock &lock)
    {
       CHECK_STATIC_INIT();
       
       CXXTLS_HEAVYDEBUG("Begin CV wait: " << this);


#       ifndef _MSC_VER
            int err = pthread_cond_wait(&condition_, &lock.mutex()->mutex());
        
            if(err)
            {
               throw Mutex::ConditionVariableWaitFailure();
            }
#       else
            SleepConditionVariableCS(&condition_, &lock.mutex()->mutex(), INFINITE);
#       endif

        CXXTLS_HEAVYDEBUG("END CV wait: " << this);

    }

    void ConditionVariable::notifyOne()
    {
       CHECK_STATIC_INIT();

#       ifndef _MSC_VER

            int err = pthread_cond_signal(&condition_);
           
            if(err)
            {
                throw Mutex::ConditionVariableNotifyFailure();
            }
#       else

            WakeConditionVariable(&condition_);

#       endif

        CXXTLS_HEAVYDEBUG("Notify one CV waitor: " << this);

    
    }


#if 0
    bool ConditionVariable::waitUntil(ScopedMutexLock &lock, const struct timespec &abstime)
       /// Same as wait() above but will terminate automatically a specified absolute
       /// time.  That is, the time specified is a specific time, not a relative time.
       /// To obtain a specific future time, call clock_gettime() and then add some
       /// offset that you wish to delay to it.
       ///
       /// @param lock    The scoped lock to associate with.
       /// @param abstime The time of day to wake up from any wait we might be doing.
       ///
       /// @return true if a timeout occurred, otherwise zero.
       ///
       /// @note An exception will be thrown if an error occurs.   @b Also, 
       ///       the function clock_gettime() is in the real time library, use -lrt
       ///       to link applications that want to use clock_gettime().
       ///
    {
#       ifndef _MSC_VER

            int err = pthread_cond_timedwait(&condition_, &lock.mutex()->mutex(), &abstime);
            
            if(err)
            {
               if(err == ETIMEDOUT)
                  return true;
            
               throw Mutex::ConditionVariableWaitFailure();
            }
            
            return false;   // no timeout

#       else
            struct timespec now = gettimeofday();

            long long delta = abstime.tv_sec - now.tv_sec;
            
            delta *= (1000 * 1000 * 1000);
            delta += (abstime.tv_nsec - now.tv_nsec);
            delta /= (1000 * 1000 * 1000)  
            // delta is now in milliseconds

            if( 0 == SleepConditionVariableCS(&condition_, &lock.mutex()->mutex(), delta) )
            {
               if(GetLastError() == ERROR_TIMEOUT)
                  return true;

               throw Mutex::ConditionVariableWaitFailure();
            }

            return false; // no timeout

#       endif

    }
#endif


    void ConditionVariable::notifyAll()
    {
       CHECK_STATIC_INIT();

#       ifndef _MSC_VER
            
            int err = pthread_cond_broadcast(&condition_);
            
             if(err)
             {
                 throw Mutex::ConditionVariableNotifyFailure();
             }
#       else
             WakeAllConditionVariable(&condition_);
#       endif

        CXXTLS_HEAVYDEBUG("Notify all CV waitors: " << this);

    }

    ConditionVariable::~ConditionVariable()
    {
#      ifndef _MSC_VER
           pthread_cond_destroy(&condition_);
#      else
           // nothing needed windows?  Seems odd but the web pages don't show anything.
#      endif
    }


}; // namespace cxxtls
