#ifndef STRINGEXPRESSION_H_included_p
#define STRINGEXPRESSION_H_included_p
//
// Copyright 2012-2016, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file stringexpression.h
///
///  This file defines classes, namespaces, and functions used to 
///  executeHelper command line string expressions like those found in bash scripts.
///
///  You can use 'man bash' on a unix/linux system and look for the 
///  section called 'EXPANSION, to see the full complexity of bash
///  string expressions.  
///
///  If you need to also PARSE such strings, you need to include
///  stringexpressionparser.h.  Note that if you include that file, you 
///  automatically include this file, but not vice versa.
///
///  struct cxxtls::StringExpression holds a representation of a string expression
///  and the function, cxxtls::StringExpr::parse(), parses string expressions out
///  of a stream of characters.
///
///  StringExpressions are representations of the components of an
///  expression that when executeHelperd will give you a string.  To executeHelper
///  an expression, you must supply an expression context which defines
///  environment variables, command line arguments, and execution
///  options.  See cxxtls::StringExpr::ExecutionContext.
///

#include <cxxtls/patternmatch.h>
#include <cxxtls/file.h>
#include <cxxtls/stringarray.h>
#include <cxxtls/threadSafeHandle.h>
#include <cxxtls/exception.h>
#include <cxxtls/foreach.h>

#include <algorithm>
#include <string>
#include <list>
#include <vector>
#include <map>
#include <ctype.h>
#include <memory>
#include <portable_io.h>
#include <stdio.h>

namespace cxxtls
{

namespace ScriptFunctions
{
   // forward declaration

   struct Statement;
}


struct StringExpression;

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
namespace StringExpr
   /// implementation details for the StringExpression class.
{
    DECLARE_CXXTLS_EXCEPTION(InvalidVariableType);  
    DECLARE_CXXTLS_EXCEPTION(AssociativeArraysCannotBeSetToStringValues);


    //==============================================================
    class VariableValue
       /// An object that represents a variable in a script language
       /// interpreter.  In this implementation only strings are
       /// supported.  They can be marked for export, or not.  Export
       /// means that they will be passed to child processes.  By default,
       /// variables are not passed.  You must call the export function
       /// to make them exportable.
       ///
       /// @b Warning:  do not derive from this class unless you change the operator=()
       /// so that it does not destruct *this and copy from rhs.
    {
    public:

        typedef std::map< std::string, std::string > MapType;

        enum Type  // what kind of variable is this?  Yea, this is a gross way to do it
        {          // but the number of types is not likely to grow more than by 1 -- to include
                   // structures if I ever get to it.

            // variable assigments are implemented in cxx/ScriptLanguage/scriptfunctions.cxx in VariableAssignmentStatement::execute
            // array and map index expressions are evaluated in cxx/ScriptLanguage/stringexpression.cxx in VariableExpansionNode::execute

            vvTypeString=0,    
            vvTypeArray=1,
            vvTypeMap=2,
            vvTypeInteger=3,

            vvTypeUnknown=16  // must be last.  the number must be unique, but otherwise means nothing
        };

    private:

        union   // yeah, I know, this is annoying -- but the total number of types is small
                // and will not grow
        {
            std::string           *stringValue_;
            cxxtls::StringArray   *arrayValue_;
            MapType               *mapValue_;
            int                    intValue_;
        };

        Type type_;  
        bool exported_;

    public:

        Type type() const { return type_; }   // get the variable type

        char const *typeName() const 
           /// @eturn the name of the variable type.  All the variable type names are the same width.
           /// to simplify formatting.
        {
           // yeah, I know this is gross -- but the number of types is not likely to grow
            switch(type_)
            {
               case vvTypeString:  return "string ";
               case vvTypeArray:   return "array  ";
               case vvTypeMap:     return "map    ";
               case vvTypeInteger: return "integer";
               default:
                   break;
            }

            return "unknown";
        }

        ~VariableValue()  // destructors are not allowed to throw -- just so you know
        {
           // yeah, I know this is gross -- but the number of types is not likely to grow

           switch(type_)
           {
              case vvTypeString:   delete stringValue_; break;
              case vvTypeArray:    delete arrayValue_;  break;
              case vvTypeMap:      delete mapValue_;    break;
              case vvTypeInteger: 
                  break;

              // no can do this big guy:  default:             throw InvalidVariableType(); break;
                 
           }
        }

        VariableValue()
        :  exported_(false)
        ,  type_(vvTypeString)
        {
            stringValue_ = new std::string();
        }

        VariableValue(std::string const &v, bool exported=false)
        :  exported_(exported)
        ,  type_(vvTypeString)
        {
            stringValue_ = new std::string(v);
        }

        VariableValue(char const *v, bool exported=false)
        :  exported_(exported)
        ,  type_(vvTypeString)
        {
            stringValue_ = new std::string(v);
        }

        explicit VariableValue(Type t, bool exported=false)
        :  exported_(exported)
        ,  type_(t)
        {
           // yeah, I know this is gross -- but the number of types is not likely to grow

            switch(t)
            {
                default:             throw InvalidVariableType(); break;

                case vvTypeString:  stringValue_ = new std::string(); break;
                case vvTypeArray:   arrayValue_  = new cxxtls::StringArray(); break;
                case vvTypeMap:     mapValue_    = new MapType(); break;
                case vvTypeInteger: intValue_    = 0; break;

            }
        }


        VariableValue(VariableValue const &rhs)
        :  exported_(rhs.exported_)
        ,  type_(rhs.type_)
        {
           // yeah, I know this is gross -- but the number of types is not likely to grow

           switch(type_)
           {
              case vvTypeString:   stringValue_ = new std::string(*rhs.stringValue_); break;

              case vvTypeArray:    
              {
                  arrayValue_  = new cxxtls::StringArray();  

                  for(size_t i = 0; i < rhs.arrayValue_->size(); ++i)
                      arrayValue_->push_back((*rhs.arrayValue_)[i]);
              }
              break;
              case vvTypeMap:      mapValue_    = new MapType(*rhs.mapValue_);    break;
              case vvTypeInteger:  intValue_    = rhs.intValue_;                  break;

              default:             throw InvalidVariableType(); break;
                 
           }
        }

        std::string value() const    // return a string not a reference due to thread safety and other issues
        {

           // yeah, I know this is gross -- but the number of types is not likely to grow

           switch(type_)
           {
              case vvTypeInteger:  return cxxtls::lexical_cast<std::string>(intValue_); 
              case vvTypeString:   return *stringValue_;
              
              
              case vvTypeArray:    
              {
                 return (*arrayValue_)[0];  // StringArray will return "" if no such item exists

              }
              break;

              case vvTypeMap:      
              {
                  MapType::const_iterator location = mapValue_->find("0");  // to keep in sync with declare -a though it makes little sense

                  if(location != mapValue_->end())
                     return location->second;

                  return std::string();

              }
              break;

              default:             throw InvalidVariableType(); break;
                 
           }

           // should never get here


        }

        size_t arrayLength() const
            /// get the length of this array variable -- or throw an exception.
        {
           if(type_ != vvTypeArray)
           {
               throw InvalidVariableType();    
           }

           return arrayValue_->size();

        }

        size_t mapLength() const
            /// get the length of this array variable -- or throw an exception.
        {
           if(type_ != vvTypeMap)
           {
               throw InvalidVariableType();    
           }

           return mapValue_->size();

        }



        void setArrayValue(StringArray const &rhs)
        {
           if(type_ != vvTypeArray)
           {
               throw InvalidVariableType();    
           }

           arrayValue_->clear();

           size_t count = rhs.size();

           for(size_t i = 0; i < count; ++i)
               arrayValue_->push_back(rhs[i]);
        }


        void setValue(std::string const &rhs)
        {
           // yeah, I know this is gross -- but the number of types is not likely to grow

           switch(type_)
           {
              case vvTypeInteger:  
              try
              {
                  intValue_ = cxxtls::lexical_cast<int>(rhs);
              }
              catch (...)
              {
                  intValue_ = 0;
              }
              break;

              case vvTypeString:   stringValue_ = new std::string(rhs);         break;
              
              case vvTypeArray:    
              {
                  arrayValue_->clear();
                  arrayValue_->push_back(rhs);
              }
              break;

              case vvTypeMap:      
              {
                  throw AssociativeArraysCannotBeSetToStringValues(); 
              }
              break;

              default:             throw InvalidVariableType(); break;
                 
           }

           // should never get here
            
        }  // setValue()

        std::string arrayDereference(int index) const
        {
           // yeah, I know this is gross -- but the number of types is not likely to grow

           switch(type_)
           {
              case vvTypeArray:    
              {
                  if(index >= arrayValue_->size())
                     return std::string();

                  return (*arrayValue_)[index];
              }
              break;

              default:             throw InvalidVariableType(); break;
                 
           }

           // should never get here

        }  // arrayDereference

        std::string mapDereference(std::string const & key) const
        {
           // yeah, I know this is gross -- but the number of types is not likely to grow

           switch(type_)
           {
              case vvTypeMap:    
              {
                  MapType::const_iterator where;

                  where = mapValue_->find(key);

                  if(where == mapValue_->end())
                     return std::string();

                  return where->second;

              }
              break;

              default:             throw InvalidVariableType(); break;
                 
           }

           // should never get here

        }  // mapDeference

        MapType::const_iterator mapBegin() const { return mapValue_->begin(); }  // for iterating over map elements
        MapType::const_iterator mapEnd() const { return mapValue_->end(); }


        void setLocation(int index, std::string value)
        {
            if (type_ != vvTypeArray)
                throw InvalidVariableType();

            (*arrayValue_)[index] = value;   // StringArray creates missing cells as necessary
        }

        void setLocation(std::string const &key, std::string value)
        {
            if (type_ != vvTypeMap)
                throw InvalidVariableType();

            MapType::iterator location = mapValue_->find(key);

            if (location != mapValue_->end())
            {
                location->second = value; // replace existing value
            }
            else
                (*mapValue_)[key] = value;  // ad new one

        }


        void setExportFlag(bool exported)
        {
           exported_ = exported;
        }

        bool exportFlag() const
        {
           return exported_;
        }

        operator std::string () const { return value(); }

        VariableValue &operator=(VariableValue const &rhs)
        {
            // I really should swap the implementations of copy constructor and this function
            // but the copy constructor basically deletes the thing anyway, so it isn't really
            // much difference.  Only if the assignment were of the same types would this not be a destroy followed
            // by a reconstruct.

            this->VariableValue::~VariableValue();  // destruct *this

            new (this) VariableValue(rhs);  // copy construct
            return *this;
        }

        VariableValue &operator=(std::string const &rhs)
        {
            setValue(rhs);
            return *this;
        }

        VariableValue &operator=(char const *rhs)
        {
            setValue(rhs);
            return *this;
        }

        typedef std::string::iterator       iterator;
        typedef std::string::const_iterator const_iterator;
        typedef std::string::value_type     value_type;

#if 0
        const_iterator begin() const { return value_.begin(); }
        const_iterator end  () const { return value_.end(); }

        iterator begin() { return value_.begin(); }
        iterator end  () { return value_.end(); }
#endif



        size_t size() const 
           /// @return the number of bytes that it takes to hold a human readable representation
           ///  of the string.  Note that the return value of this function is used to allocate
           ///  space when invoking child processes so it it must give a value of the ascii
           ///  representation of string's size.  For integers, maps, and arrays, the 
           ///  function must convert the variable to a string, using the value() function
           ///  and return the size thereof.
        { 
            switch(type_)
            {
               case vvTypeString:   return stringValue_->size();  break;
               default:
                  break;
            }

            std::string text = value();  // converts arrays and maps to a string

            return text.size();


        }






        bool   empty() const 
        { 
            switch(type_)
            {
               case vvTypeString:   return stringValue_->empty();  break;
               case vvTypeArray:    return arrayValue_->empty();   break;
               case vvTypeMap:      return mapValue_->empty();   break;
               default:             throw InvalidVariableType(); break;
            }

            // you should never get here
        }
        
    };

    //==============================================================
    struct ExecutionVariables
        /// The object that holds the normal environment variables for
        /// an execution context.  
        ///
        /// An ExecutionVariables object does
        /// not exist in the ExecutionContext itself so as to allow
        /// the variable superseeds in an efficient way.  When a 
        /// bash command that looks like this is executed:
        ///
        ///@begincode
        ///
        ///   fred=bill tom=sue  CommandName parm1 parm2 parm3 ...  
        ///
        ///@endcode
        ///
        /// Then CommandName is executed using a copy of the current
        /// base environment variables with fred and tom superseeded
        /// as specified.  If the ExecutionContext were to hold a copy
        /// of the environment variables, then a whole copy of the
        /// execution context would have to be made in order to 
        /// pass the modified variables to CommandName.  But with the
        /// ExecutionVariables in a separate structure, the ExecutionContext
        /// can be passed quickly, along side a tiny "superseed" structure
        /// that holds the modified variables.  
        ///
        /// The ExecutionContext structure holds a pointer to an ExecutionVariables
        /// object which is actually stored in the over all interpreter's "context".
        /// ExecutionContext also holds an ExecutionVariables object which is named
        /// superseeds_ and it only contains the current command line superseeded
        /// variables, see above.
        /// 
    {
       typedef std::vector<std::string>                                             ArgsVector;
       typedef std::map<std::string,VariableValue>                                  VariablesMap;
       typedef std::map<std::string, ArgsVector>                                    AliasMap;

       struct FunctionInfo
           // the location a function was defined and the statement that comprises its body.
       {
           ThreadSafeHandle<ScriptFunctions::Statement> statement_;
           std::string                                  file_;
           int                                          line_;

           FunctionInfo(ThreadSafeHandle<ScriptFunctions::Statement> &statement,
                        std::string const &file,
                        int                line
                       )
           :   statement_(statement)
           ,   file_(file)
           ,   line_(line)
           {
           }

           FunctionInfo()
           {
              file_ = "";
              line_ = 0;
           }
       };

       typedef std::map<std::string, FunctionInfo > FunctionsMap;



       bool contains(std::string const &name) const
          /// @return true if the this set of variables contains the named item.
       {
           return vars_.find(name) != vars_.end();
       }


       class ProcessVariableSet
          /// a wrapper around the data needing to be fed to the process creation routines
          /// to encapsulate its environment.  
          ///
          /// The implementations are different on windows and unix, but this class provides a
          /// wrapper around the difference.
          ///
          /// Call  ExecutionVariables::getProcessVariables(someProcessVariableSetObject) to
          /// populate it.  Then call member vars() to use the data.
       {
#ifdef _MSC_VER

          private:

              std::vector< char > vars_;   /// see windows CreateProcess() documentation for what the content
                                           /// must be like.

              friend class ExecutionVariables;

          public:

          char *vars() { 
                          static char empty[5] = { 0, 0, 0, 0, 0 };  // an empty process environment.  should never happen.

                          if(vars_.empty())
                              return empty;

                          return &vars_[0]; 
                       }
#else

#endif
          
       };

       void getProcessVariables(ProcessVariableSet &vars) const;
           /// Populate a set of process variables (environment variables) from this ExecutionVariables
           /// object.  The implementaton is different on unix and windows


       void set(std::string const &name, std::string const &value)
           /// define an environment variable by giving it a name
           /// and a value.  You can change the variable using
           /// this same function.
            {
                VariablesMap::iterator location = vars_.find(name);

                if(location != vars_.end())
                {
                   location->second = value;
                }
                else
                {
                   vars_[name]=value;
                }

            }

       void createVariable(std::string const &name, VariableValue::Type type)
           /// define an environment variable by giving it a name
           /// and a type.  The value will be empty for the given type.
           /// variable types are string, integer, array and map.
            {
                VariablesMap::iterator location = vars_.find(name);

                if(location != vars_.end())
                {
                    if(type == location->second.type())
                    {
                       // existing variable of correct type

                       return;  // nothing to do
                    }

                    // extisting variable of different type

                    vars_.erase(location);

                }

                // at this point, no variable by the name exists

                vars_[name] = VariableValue(type);
            }

       void createIntegerVariable(std::string const &name)
       {
          createVariable(name, VariableValue::vvTypeInteger);
       }

       void createArrayVariable(std::string const &name)
       {
          createVariable(name, VariableValue::vvTypeArray);
       }

       void createStringVariable(std::string const &name)
       {
          createVariable(name, VariableValue::vvTypeString);
       }

       void createMapVariable(std::string const &name)
       {
          createVariable(name, VariableValue::vvTypeMap);
       }

       VariableValue::Type variableType(std::string const &name) const
       {
           VariablesMap::const_iterator location = vars_.find(name);
           
           if(location != vars_.end())
              return location->second.type();

           return VariableValue::vvTypeUnknown;
       }


       bool erase(std::string const &name)
           /// delete a variable by name
       {
                VariablesMap::iterator location = vars_.find(name);

                if(location != vars_.end())
                {
                   vars_.erase(location);
                   return true;
                }

                return false;
       }

       std::string get(std::string const &name) const
           /// find the value associated with an environment variable
           /// given the name of the variable.
           ///
           /// If the variable does not exist, an empty string will
           /// be returned.
       {
           VariablesMap::const_iterator location = vars_.find(name);
       
           if(location != vars_.end())
           {
              return location->second;
           }
       
           return std::string();
          
       }

       bool exportFlag(std::string const &varname) const
       {
           VariablesMap::const_iterator location = vars_.find(varname);
       
           if(location != vars_.end())
           {
              return location->second.exportFlag();
           }
           return false;
       }

       void setExportFlag(std::string const &varname, bool value)
       {
           VariablesMap::iterator location = vars_.find(varname);
       
           if(location != vars_.end())
           {
              location->second.setExportFlag(value);
           }
           else
           {
               set(varname,"");
               setExportFlag(varname, value);  // could recur infintely if set fails...

           }
       }



       typedef VariablesMap::const_iterator  const_iterator;
       typedef VariablesMap::iterator        iterator;
       typedef VariablesMap::value_type      value_type;

       const_iterator begin() const { return vars_.begin(); }
       const_iterator end()   const { return vars_.end(); }

       const_iterator find(std::string const &name) const { return vars_.find(name); }
       iterator       find(std::string const &name)       { return vars_.find(name); }


       void mergeFrom(ExecutionVariables const &superseeds)
          /// Incorporate changes made in another variables list.  The superseeds take precedence.
          ///
          /// This function is called by unifyVariables for the purpose of creating a clone of the
          /// existing environment.  The primary purpose of said clone is to launch child jobs
          /// that are running in a separate thread.
          ///
          /// This function makes sure that the strings in the merged copy are unique to the 
          /// ExecutionVariables object so that they will be unrelated by any std::string magic
          /// to strings in other threads.  It does this by creating new strings without using
          /// the std::string copy or assign functionality.
          ///
          /// See unifyVariables for other usage notes about this function
          /// 
       {

          const_iterator first = superseeds.begin(), last = superseeds.end();

          while(first != last)
          {
             std::string const name(first->first.begin(), first->first.end());  // copy name to new string

             std::string varValue = first->second.value();

             std::string const value(varValue.begin(), varValue.end());  // copy value to new string, byte by byte in a
                                                                         // thread safe manner to prevent two threads
                                                                         // pointing to the same hidden string magic data

             iterator outputPtr = vars_.find(name);

             if(outputPtr != end())
             {
                 outputPtr->second = value; // value is a unique block of characters at this point with no
                                            // std::string magic references to strings in other threads
             }
             else
             {
                 set(name,value);
             }

             ++first;
          }

       }


       void includeProgramEnvironment();
           /// add/replace variables in this object with the contents of the program
           /// global environment.  This should only be done once at the begining of an
           /// application run.

       void clear() { vars_.clear(); } ///<  get rid of all variables


       typedef VariablesMap::iterator        iterator;

       iterator begin() { return vars_.begin(); }
       iterator end() { return vars_.end(); }

       void clearExecHash()
       {
          execHash_.clear();
       }

       std::map<std::string, std::pair<std::string,int> > &execHash() 
       {
          return execHash_;
       }

       enum execHashEntity
          ///  a list of the types of executables that can be kept track of
       {
          unknownEntityType     = -1,
          nativeExecutable      = 0,
          defaultScriptLanguage = 1,
       };

       std::pair<std::string,int> findInExecHash(std::string const &programName, int hashEntitytype );
           ///<  given program name, if it has no path component, find it in
           ///   the PATH variable and add it as a mapping between the nodename
           ///   and the full pathname.
           ///
           ///   @param programName -- the name of a program, with or without a directory path
           ///                         that defines a program to be executed.
           ///
           ///   @param hashEntityType -- a definition of what kind of executable the thing
           ///                           is -- 0 is native program, 1 is "script of the current language"
           ///                           and other integer  values are scripts of other languages --
           ///                           based on the executable into which this code is linked.
           ///   @return  A pair consisting of a string and an integer that defines the following:
           ///
           ///     @li  the full pathname that the program name refers to if a $PATH search was
           ///          performed
           ///
           ///     @li  the entity type of the item in the execution hash table.  The types are as
           ///          follows:
           ///          <ul>
           ///            <li> unknown:  -1
           ///            <li> native executable:  0
           ///            <li> default script language of this program: 1
           ///          </ul>

       void addToExecHash(std::string const &commandName, std::string const &programPath, int hashEntitytype);
           ///<  Given a command name, without a path or extension, and a program name or script file name
           ///   that implements the command, insert a hash map entry
           ///
           ///   @param commandName, see above.
           ///   @param programPath, see above.
           ///   @param hashEntityType, see findInExecHash

       const_iterator findIterator(std::string const &name) const
       {
          return vars_.find(name);
       }


       AliasMap const &aliases() const              { return aliases_; }  ///<  get the list of aliases

       void addAlias(std::string const &name, ArgsVector const &contents) 
           ///  Add a new alias
       {
          AliasMap::iterator where = aliases_.find(name);

          if(where == aliases_.end())
             aliases_[name] = contents;
          else
             where->second = contents;
       }

       void removeAlias(std::string const &name)
           /// remove a named alias
       {
          AliasMap::iterator where = aliases_.find(name);

          if(where != aliases_.end())
              aliases_.erase(where);
           
       }

       void removeAllAliases()
       {
           aliases_.clear();
       }


       void addFunction(std::string const &name, ThreadSafeHandle<ScriptFunctions::Statement> &f, std::string const &file, int line)
       {
           // the following insert causes the std::string holding the name in the map to be unique across all threads.

           FunctionsMap::iterator where = functions_.find(name);

           if(where != functions_.end())
           {
              functions_.erase(where);
           }

           FunctionInfo fInfo(f, file, line);


           functions_.insert( std::pair<std::string, FunctionInfo >(name.c_str(), fInfo) ); 

       }

       FunctionsMap const &functions() const { return functions_; } 

       void removeFunction(std::string const &name)
       {
           FunctionsMap::iterator where = functions_.find(name);

           if(where != functions_.end())
              functions_.erase(where);
       }

       FunctionInfo  findFunction(std::string const &name) const
       {
          FunctionsMap::iterator where = functions_.find(name);

          if(where != functions_.end())
             return where->second;

          return FunctionInfo();  // return null pointer, basically
       }


       void setExecLog(bool on) { execLog_ = on; }
       bool execLog() const     { return execLog_; }

       ExecutionVariables()
       :  execLog_(false)
       {
          CXXTLS_DEBUG("Constructing ExecutionVariables(" << (void*)(this) << ")");
       }

       ~ExecutionVariables()
       {
          CXXTLS_DEBUG("~ExecutionVariables(" << (void*)(this) << ")");
       }

       private:

           VariablesMap        vars_;

           std::map<std::string, std::pair<std::string,int> > execHash_; // the mapping between commands with no pathname and their
                                                                         // full pathname.

           AliasMap            aliases_;

           mutable FunctionsMap        functions_;

           bool execLog_; // not used in this code, but this is the most convenient place for the flag
    };  // class ExecutionVariables

   struct LoggerBase
   {
       virtual void log(std::string const &message) const=0; // call this function to perform a logging operation
   };


    //==============================================================
    struct ExecutionContext
       /// A partial representation of a bash script's execution state:
       /// 
       /// @li a pointer to the environment variables for the current interpreter context 
       /// @li a vector of command line arguments
       /// @li various option settings 
       /// @li command line superseeds to the environment variables
       /// @li aliases maintained by the script language parser.
       ///
       /// Intentionally absent from the ExecutionContext is a
       /// container of functions.
       /// 
       /// But it does have the alias map.  An alias is a name that is intepreted by the script language
       /// parsers to identify text fragments that should be interpreted.  Basically, like parameter-less
       /// macros.
       ///
       /// When accessing variable values, it is generally preferable to call findVariableOrValue()
       /// which will handle variables, superseeds, and command line arguments.  Command line
       /// arguments have names like $@, $1, $0, etc.  For normal variable names, it looks first a 
       /// superseeds then at base values for environment
       /// variable values.
       ///
       /// To set normal variables, use the assign() function.  To set a superseed, use superseed().
       ///
       /// When you set a normal variable value, using assign(), it clears out any superseed by that
       /// name and then sets the variable in the current environment.
       ///
       /// The 'unifyVariables()' function lets you make a merger of the two
       /// as when you need a subshell or if you are running a program.
       ///
       /// Note that the interface to command line arguments has a special feature:
       ///
       ///   @li Note only can you get and set arguments, you can swap the entire sets
       ///   of arguments out with a different set in a single function call. See
       ///   swapArguments.
       ///
       /// The purpose of this feature is to allow for very quick and temporary replacement
       /// of the arguments when you are calling a function.  The called function can use
       /// the same arguments as its parent scope but the code that implements the function
       /// will temporarily swap the functions arguments with the arguments in the ExecutionContext.
       /// At exit, the caller's arguments must be swapped back into place.  This is much
       /// faster than creating a stack of vector of string.  But carries the programming responsibility
       /// of the function call handler not to mis-place the caller's arguments...
       ///
    {

       typedef ExecutionVariables::ArgsVector  ArgsVector;
       typedef ExecutionVariables::AliasMap    AliasMap;

       ExecutionContext(ExecutionVariables *variables) 
       : vars_(variables)
       , logging_(false)
       , errorMode_(false)
       , errorCode_(0)
       , sepChar_(' ')
       , threadExecutionContext_(0)
       , expandFilenames_(true) // use set -f to turn it off, and +f to turn it on
       , logger_(0)             // function that can perform the logging operation
                                // in the few cases where statements are executed directly in the ExpressionContext class
       {
           wordSplitter_.setIFS(variables->get("IFS"));
       }



       char sepChar() const { return sepChar_; }  ///< space or first char of IFS


       virtual ~ExecutionContext() 
       {
          CXXTLS_DEBUG("~ExecutionContext(" << (void*)(this) << ")" ) ;

          // so derived classes can be treated like this class but have more variables -- such as
          // a private copy of the ExecutionVariables, etc.
       }

       void setLogging(bool flag) { logging_ = flag; }  ///< turn logging on or off
       bool logging() const { return logging_; } ///< get the logging flag

       void setGlobbing(bool flag) { expandFilenames_ = flag; } ///<  turn glob on or off.

       bool globbing() const { return expandFilenames_; } ///<  get the glob flag

       void unifyVariables(ExecutionVariables &output) const
          /// merge my input variables and my output variables and store them in an output 
          /// ExecutionVariables object for use as the basis of a sub-shell used by some other
          /// command.
       {
           output.clear();  // we are going to create a thread safe copy of the data of interest
           output.mergeFrom(*vars_);
           output.mergeFrom(superseeds_);
       }


       void superseed(std::string const &name, std::string const &value)
          ///  In order to support prefix variable assignments to bash commands, this function
          ///  lets you add things to the superseeded variable values.  Here's the bash code that
          ///  would use this function:
          ///
          ///@begincode
          ///
          ///  FRED=stuff SUE=more CommandName parm1 parm2 ...
          ///
          ///@endocode
          ///
          /// In the above example, FRED and SUE are superseeds of the environment variables by the same
          /// name and are applicable only to the invocation of CommandName this one time.
       {
           superseeds_.set(name,value);
       }

       std::string find(std::string const &varname) const
           ///  get a variable's value -- either from the superseed or from the parent context's variables.
           ///  
           ///  Don't use this function to deal with $@, $*, $1 etc.  Instead, use findVariableOrValue()
           ///  which handles normal variables by calling find() itself, or by interpreting the names *,
           ///  @, 1, 2, etc.
       {

           if(superseeds_.contains(varname))
           {
               return superseeds_.get(varname);
           }

           ExecutionVariables::const_iterator loc = vars_->findIterator(varname);

           if(loc == vars_->end())
           {
               // there is no mapping for this variable

               if(undefinedVariablesHandlers_)
               {
                   UndefinedVariablesHandlers::const_iterator where = undefinedVariablesHandlers_->find(varname);

                   if(where != undefinedVariablesHandlers_->end())
                   {
                       return (*where->second)(this);
                   }

                   return std::string();

               }

           }
           else
               return loc->second;

           return std::string();  // nothing

       }  // find


       void assign(std::string const &name, std::string const &value)
          /// perform a user visible variable assignment.  The set function, described above only modifies
          /// the superseeds to values wrapped by this execution context.  If the user writes code like the
          /// following:
          ///
          ///@begincode
          ///
          ///  NAME=value
          ///
          ///@endcode
          ///
          /// Then the user is trying to set the variable, NAME, to a value in the actual script context.
          ///
          /// This is the function that does that
       {
          if(name.empty())
              return;  // could throw.  should we?

           superseeds_.erase(name);   // get rid of the superseed.  Once a shell function changes the
                                      // value, the superseed vanishes magically.

           vars_->set(name,value);  // set it in the environent too

           if(name == "IFS")
           {
               wordSplitter_.setIFS(value);

               if(value.size())
                  sepChar_ = value[0];  // set up for the separator character

           }
       }

       void assign(std::string const &name, StringArray const &value)
          /// perform a user visible variable assignment of an array value.  
          ///
          ///@begincode
          ///
          ///  NAME=(value1 value2 value3) ...
          ///
          ///@endcode
          ///
          /// Then the user is trying to set the variable, NAME, to an array value in the actual script context.
          ///
          /// This is the function that does that
       {
          if(name.empty())
              return;  // could throw.  should we?

           superseeds_.erase(name);   // get rid of the superseed.  Once a shell function changes the
                                      // value, the superseed vanishes magically.

           vars_->erase(name);

           vars_->createArrayVariable(name);

           ExecutionVariables::iterator location = vars_->find(name);

           location->second.setArrayValue(value);

       }



       void addArgument(std::string const &p)
       {
          arguments_.push_back(p);
       }

       void swapArguments(ArgsVector &newbie, std::string &newScriptName)
           ///  Replace all the arguments with those found in a different set.  This is 
           ///  dramattically faster than an assignment, but replaces the source (newbie)
           ///  with the old contents.
           ///
           ///  This function is useful for handling the arguments to functions.  While
           ///  executing a function, the function's arguments override the calling environt's
           ///  args.  So the caller can compute the functions arguments, swap them in the
           ///  environment, then on exit, swap them back to get the originals.
       {
           arguments_.swap(newbie);
           scriptName_.swap(newScriptName);
       }

       ArgsVector const &arguments() const  { return arguments_; } ///< get command line arguments

       std::string const &scriptName() const { return scriptName_; } ///< get script name

       void setArguments(ArgsVector const &v) 
            /// set the command line args by cloning them in a thread safe way -- that is,
            /// copy the bytes in the arguments, not just the magic std::string references.
       { 
           arguments_.clear();

           ArgsVector::const_iterator f = v.begin(), l = v.end();

           while(f != l)
           {
              std::string const &cur = *f++;

              std::string tmp(cur.begin(), cur.end());

              arguments_.push_back(tmp); 
           }

       }  


       void shiftArguments(size_t shiftCount)
       {
           if(arguments_.size() >= shiftCount + 1)
           {
              ArgsVector::iterator firstVictim = arguments_.begin() + 1;
              ArgsVector::iterator lastVictim  = firstVictim + shiftCount;   // out side the range, as normal

              arguments_.erase(firstVictim, lastVictim);
           }
           else
           {
              arguments_.erase( arguments_.begin() + 1, arguments_.end());  // nuke everything but the script name
           }
       }

       void setScriptName(std::string const &n) 
           /// set the script name and make it an atomic copy
       { 
           scriptName_ = std::string(n.begin(), n.end());   // a threadsafe copy
       } 

       int errorCode() const { return errorCode_; } ///< get the most recent error code

       void setErrorCode(int i) { errorCode_=i; } ///< set the most recent error code value

       virtual void findVariableOrValue(std::string const &name, std::string &rv) const ;
            /// Evaluate variables given a name.  The name can be a string, as one would
            /// use when defining a named variable.  The following other strings can be used
            /// as well:
            ///
            ///    0  -- gives you the script name 
            ///
            ///    1  -- gives you the 1'th script parameter
            ///    ...
            ///    9  -- gives you the 9'th script parameter
            ///
            ///    #  -- gives you the count of script parameters
            ///
            ///    ?  -- gives you the error code from the context
            ///
            ///  Note: undefined variables are checed against the undefinedVariablesHandlers_ table to
            ///  see if there is an application defined mapping for the named variable

       ExecutionVariables &variables() { return *vars_; }  ///<  get the variables for iteration purposes, does not
                                                           ///   include the superseeds

       void setExportFlag(std::string const &varname, bool flag)
          /// control the export flag associated with a variable
       {
           vars_->setExportFlag(varname, flag);
       }


       bool errorMode()            const { return errorMode_; }  // use set -e to turn it off in the script language
       void setErrorMode(bool v)         { errorMode_ = v;    }  

       ExecutionVariables const &superseeds() const { return superseeds_; }

       ExecutionVariables::AliasMap const &aliases() const { return vars_->aliases(); }  ///<  get the list of aliases

       void addAlias(std::string const &name, ArgsVector const &contents) 
           ///  Add a new alias
       {
           vars_->addAlias(name, contents);
       }

       void removeAlias(std::string const &name)
           /// remove a named alias
       {
           vars_->removeAlias(name);
       }

       void removeAllAliases()
       {
           vars_->removeAllAliases();
       }

       void removeVariable(std::string const &varname)
       {
           superseeds_.erase(varname);
           vars_->erase(varname);
       }

       typedef std::string (*UndefinedVariableMappingFunction)(ExecutionContext const *);  // used to map SECONDS into 12345678 or whatever

       typedef std::map<std::string, UndefinedVariableMappingFunction> UndefinedVariablesHandlers;

       static void registerUndefinedVariablesHandlers(UndefinedVariablesHandlers *handlers)
          // a non-threadsafe function which lets you set up the undefined variables handlers for this 
          // implementation.  You MUST call this before multi-threading starts from your main program or whererever
          // is appropriate and the table must be constant all during the threaded phase of execution.
       {
           undefinedVariablesHandlers_ = handlers;
       }

       ExecutionVariables::FunctionInfo  findFunction(std::string const &name) const
       {
          return vars_->findFunction(name);
       }

       void addFunction(std::string const &name, ThreadSafeHandle<ScriptFunctions::Statement> &func, std::string const &file, int line)
       {
          return vars_->addFunction(name, func, file, line);
       }


       class WordSplitter
          /// a wordSplitter is an object that can split strings into pieces based on
          /// delimiter characters.
          ///
          /// Per the bash manual page, if the IFS variableis set to <space><tab><newline>
          /// or if it is not set at all, then the following splitting operation applies:
          ///
          /// 1.  leading and trailing repetitions of <space> || <tab> || <newlines> get removed
          /// 2.  groups if either <space><tab><newline>, in the middle of the text, get treated as a single delimiter.
          ///
          /// Any other setting of the IPS will result in all members of the IFS being treated as delimiters
          /// and any single instance results in a split at that character position.
          ///
          /// otherwise, the characters in IFS are treated a single character delimiters
       {
            bool wordSplitting_;  // if not word splitting, then we are splitting based on any character in the ifs.
          
            std::string ifs_;     

       public:
            WordSplitter()
            :  wordSplitting_(true)
            , ifs_(" \t\n")
            {
            }

            void split(std::string const &input, std::vector< std::string > &output) const;  // peform the split
               /// see class comments

            void setIFS(std::string const &ifs);
               /// see class comments

            bool operator()(char c) const
               ///  A WordSplitter can be used as a boolean functor that tells you if a character is a delimiter
            {
               if(ifs_.find_first_of(c) < ifs_.size())
                  return true;

               return false;
            }

       };

       void splitWords(std::string const &input, std::vector<std::string> &output) const
          /// Split strings into words based on bash rules:
          ///   @li if IFS = <space><tab><newline>
          ///       then split on words by removing
          ///       lead, trailing  IFS contents
          ///       and treating groups of IFS
          ///       characters as a single delimiter.
          ///   @li otherwise, treat each character
          ///       in the IFS, as a separate delimiter
          ///       and all individual delimiters appearing
          ///       in the source start a new word --
          ///       which may be empty string.
          ///        
       {
          return wordSplitter_.split(input, output);
       }


       WordSplitter const &wordSplitter() const { return wordSplitter_; }


       std::string lastParmOfLastCommand() const { return lastParm_.c_str(); }  ///< for populating $_ in the expression evaluator

       void setLastParmOfLastCommand(std::string const &value);         ///< for use in populating the $_ in the expression evaluator


       int getTmpDirName(std::string &result);
           ///<  Get the name of the temp directory.  The temp directory is found like this:
           ///   @li On unix, it is either /tmp, or it is the specified by the TMP enviornment
           ///       variable.
           ///   @li On windows, it is specified by the TEMP environment variable, or it is
           ///       the home directory.  Which is specified by USERPROFILE.
           ///
           ///   @return result contains the name of the directory and the return code is one
           ///   of the following:
           ///       @li 0  for success
           ///       @li -2 means that the directory does not exist.
           ///    

       void setThreadExecutionContext(void *p) { threadExecutionContext_ = p; }
       void*   threadExecutionContext()        { return threadExecutionContext_; }

       typedef std::string (*SystemCommandExecutorFunc)(std::string const &command, void *threadExecutionContext );
           /// The signature of the function that you must define in order to superseed the default system command executor.
           ///
           /// You must also call setSystemCommandExecutor()

       static void setSystemCommandExecutor(SystemCommandExecutorFunc f ) { systemCommandExecutor_ =  f; }
           ///   overwrite the default system command executor function.

       std::string executeSystemCommand(std::string const &command)
           /// execute a system command and return the results as a string.  Note that the command will not be
           /// executed by this function, it will be sent directly to the system command invocation function which has a
           /// default value that acts like popen(), but if you superseed the command executor, using setSystemCommandExecutor(),
           /// you can provide your own function that handles the process.
           ///
           /// To make brash work, the ThreadContext constructors set their own 'this' pointer as the threadExecutionContext_
           /// member of this class, using setThreadExecutionContext(), and thus any executor function that uses ThreadContext
           /// can cast the void pointer value, threadExecutionContext_, to be a real threadExecutionContext_, and can thus
           /// use the file handles to perform io.
       {
          return (*systemCommandExecutor_)(command, threadExecutionContext());
       }

       std::string quoteChunk(std::string const &);
           ///<  Return a quoted version of the input string.  The quoting process works like this:
           ///   @li  if the input has no single quotes, then the whole mess gets returned with leading
           ///        and trailing single quotes.
           ///   @li  Otherwise, return a double quoted string with \ slashes put before the \\ and " characters.
           ///        


       void setExecLog(bool on) { vars_->setExecLog(on); }   // turn on execution logging.  This value is not used in the string expression code, it is just
                                                             // convient for it to be here.

       bool execLog() const     { return vars_->execLog(); }

    protected:

       ExecutionVariables  *vars_;       // not owned by this class!  But modified. Someone else has to clean up

       ExecutionVariables  superseeds_;  // variable superseeds only, not the main storage
       ArgsVector          arguments_;
       std::string         scriptName_;
       bool                logging_;
       bool                errorMode_;
       int                 errorCode_;

       static UndefinedVariablesHandlers *undefinedVariablesHandlers_;   // assumes not to ever change and must
                                                                         // point to table that is never modified
                                                                         // during execution

       WordSplitter        wordSplitter_;

       char                sepChar_;  // usually space, but may be first char of IFS

       std::string         lastParm_; // last parameter to the most recent command.

       void*               threadExecutionContext_;  // set by ThreadContext::create(), used only by subshellNode::executeHelper

       static SystemCommandExecutorFunc  systemCommandExecutor_; // defaults to "popen() with a bit a of wrapper around it"

       bool                execLog_;  // this is just a convenient place to put is variable

       bool                expandFilenames_;  // controlled using brash's "set -f" option

    public:

       LoggerBase *logger_;   // logs execution variables based on the logging_ flag




        void registerLogger( LoggerBase* pointer)
           // temporarily set the logger for an execution context
        {
            logger_ = pointer;
        }

        void deregisterLogger()
           // restore the unregistered state of the logger for this execution context
        {
           logger_ = 0;
        }

    };  // ExecutionContext

    struct ThreadVariablesExecutionContext
    :   public ExecutionContext
       ///  A unified copy of all the ExecutionContext components with all std::string objects being
       ///  specific to this object -- so that this is an "atomic" object from a thread safe sense.
    {
        ThreadVariablesExecutionContext(ExecutionContext const &parent)
        :  ExecutionContext(  new ExecutionVariables() )
        {
           parent.unifyVariables(*vars_);  // this creates a thread safe clone of inVars

           setArguments(parent.arguments());  // makes a thread safe copy where the memory is uniquely
                                             // referenced by vars_
        }
        
        ~ThreadVariablesExecutionContext()
        {
           delete vars_;   // we own the variables in this special case, normally, the execution
                           // context does not.
        }
    };

    //==============================================================

     struct Node
         /// A part of a string expression.  StringExpression objects
         /// contain a list of nodes that define the expression.
         /// Evaluation of the string expression concatenates the evaluation
         /// of the nodes -- possibly with decorations between them like
         /// spaces double quotes.
         ///
         /// The cxxtls::parse() method constructs a StringExpression containing
         /// the list of needed nodes.  
         ///
     {

        Node()
        {
        }

        virtual ~Node() 
        {
        }


        virtual bool allowsNoResult() const
            // in general, if a string expression exists, it will evaluate to at minium and empty
            // string as a return value.
            //
            // but for "$@", it is possible to return nothing at all, not even an empty string.
        {
           return false; 
        }

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const 
            ///  executeHelper this node in the string expression with regard to the
            ///  specified context and return true to indicate that this
            ///  expression's text fragment should be protected from globbing.
            ///
            ///  During the course of execution, any shell variable expansions
            ///  will be drawn from the context.  Command line arguments will
            ///  also be drawn from the context.  Any optional behavior
            ///  interpretations will be decided upon based on options in the
            ///  context. 
            ///
            ///  Upon completion, the result of this node's calculations will
            ///  be left in parameter rv.
            ///
            ///  Note that literal and double quoted strings will be protected
            ///  from globbing but non-protected text will not.
        { 
            output.clear();

            return false;
        }

        virtual std::string executeEncapsulated(ExecutionContext *context) const = 0;
            ///<  Evaluate the expression and return a string which is encapsulated
            ///   and encoded in a manner appropriate to the encapsulation. 
            ///
            ///   An encapsulated string is probably surrounded entirely enclosed
            ///   with double quotes, but may also be a multi-part string with
            ///   sub-components which are encapsulated and some which are not.
            ///
            ///   For example, you might get one of the following for the same
            ///   actual contents based on the original usage of the data that
            ///   becomes the string:
            ///
            ///     @li "stuff*.*"
            ///     @li "stuff"*.*
            ///     @li stuff*.*
            ///     @li 'stuff'*.*
            ///
            ///   @param[in]    context        The string execution context.
            ///
            ///   @return A string of the form:   "...", or '...', or just ...

        virtual std::string stringify() const { return std::string(); }  // debug print helper

        virtual bool isLiteral() const { return false; } // most are not, superseed only for true literals

        static  std::string encapsulateLiteralDoubleQuotes(std::string const &value)
            ///  @return an encapsulated literal string in double quotes.
        {

            std::string rv;

            rv += '"';

            std::string::const_iterator f = value.begin(),
                                        l = value.end();

            while(f != l)
            {
                char c = *f++;

                switch(c)
                {
                    // note:  newlines spaces and so forth are already handled by the fact that
                    //        we are encapsulating the output data, so no need to put in \'s for them.

                    default:  rv += c; 
                              break;

                    case '"' :  
                    case '$' :
                    case '`' :  // back tick
                              rv += '\\';   // protect the needed character
                              rv += c;
                              break;
                }
            }

            rv += '"';

            return rv;
        }

        static  std::string encapsulateLiteralNoQuotes(std::string const &value)
            ///  @return an encapsulated literal string in double quotes.
        {

            std::string rv;

            std::string::const_iterator f = value.begin(),
                                        l = value.end();

            while(f != l)
            {
                char c = *f++;

                switch(c)
                {
                    default:  rv += c; 
                              break;

                    case '"' :  
                    case '$' :
                    case '`' : // back tick
                    case '\'': // apostrophe
                    case '[' :
                    case ']' :
                    case '{' :
                    case '}' :
                    case '<' :
                    case '>' :
                    case ' ' :
                    case '&' :
                    case '?' :
                    case '*' :
                    case ';' :
                    case '(' :
                    case ')' :
                    case '\\':
                    case '\n':
                    case '\t':
                    case '\r':
                              rv += '\\';
                              rv += c;
                              break;
                }
            }

            return rv;
        }


     private:


     };

    //==============================================================

    struct Node;

    typedef std::list<Node*> Expression;
        ///<  The Node list that forms the expression.

    //==============================================================
    struct LiteralNode
        /// A literal string.  That is a sequence of characters which 
        /// have no interpretation applied to them at all.  No variable
        /// expansions, no escape sequences, etc.  A literal string is
        /// defined with a leading "'" and stops when a terminating "'"
        /// appears.  It is not possible to include a single quote in
        /// a literal node.  But you combine LiteralNodes with other
        /// types of nodes to accomplish the task.
    : public Node
    {
        std::string value_;   // 'stuff'

        bool preEncoded_; 

        LiteralNode(std::string const &s, bool preEncoded = false)
        :  preEncoded_(preEncoded)
        {
           value_.append(s.begin(), s.end());  // ensure, access across threads is safe
        }

        virtual ~LiteralNode()
        {
        }

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const 
            ///  executeHelper this node in the string expression by copying the
            ///  constant text to the output variable, rv.
            ///
            ///  No shell expandions will be performed.  Nor is there any
            ///  optional behavior.  
            ///
            ///  Note that literal and double quoted strings will be protected
            ///  from globbing but non-protected text will not.
        { 
            output.push_back(value_);

            return true;
        }

        virtual std::string executeEncapsulated(ExecutionContext *context) const
        {
            if(preEncoded_)
            {
                std::string rv;

                rv += '"';
                rv += value_;
                rv += '"';

                return rv;
            }

            return encapsulateLiteralDoubleQuotes(value_);    // not worth the bother to have a separate implementation for apostrophe

        }

        virtual std::string stringify() const { return value_; }  

        virtual bool isLiteral() const  { return true; }

    };

    ///sp
    //==============================================================
    struct SplitableLiteralNode
        /// A splitable literal string.  That is a sequence of characters which 
        /// have no interpretation applied to them at all.  No variable
        /// expansions, no escape sequences, etc.  A splitable literal string is
        /// defined by a sequence of non special characters, not including spaces,
        /// tabs or punctuation.  Not that the escape character might have been
        /// used to include other kinds of characters into the literal value
    : public Node
    {
        std::string value_;   // stuff

        bool preEncoded_; 

        SplitableLiteralNode(std::string const &s, bool preEncoded = false)
        :  preEncoded_(preEncoded)
        {
           value_.append(s.begin(), s.end());  // ensure, access across threads is safe
        }

        virtual ~SplitableLiteralNode()
        {
        }

        virtual bool executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const 
            ///  executeHelper this node in the string expression by copying the
            ///  constant text to the output variable, rv.
            ///
            ///  No shell expandions will be performed.  Nor is there any
            ///  optional behavior.  
            ///
            ///  Note that literal and double quoted strings will be protected
            ///  from globbing but non-protected text will not.
        { 
            output.push_back(value_);

            return false;
        }

        virtual std::string executeEncapsulated(ExecutionContext *context) const
        {
            if(preEncoded_)
            {
                std::string rv;

                rv += '"';
                rv += value_;
                rv += '"';

                return rv;
            }

            return encapsulateLiteralDoubleQuotes(value_);    // not worth the bother to have a separate implementation for apostrophe

        }

        virtual std::string stringify() const { return value_; }  

        virtual bool isLiteral() const  { return true; }

    };
    ///sp

    struct RawLiteralNode
    :  public LiteralNode
    {
        RawLiteralNode(std::string const &s)
        :  LiteralNode(s)
        {
        }

        virtual ~RawLiteralNode() {}

        virtual std::string executeEncapsulated(ExecutionContext *context) const
        {
                std::string rv;

                rv += '\'';
                rv += value_;
                rv += '\'';

                return rv;
        }


    };

    struct UnencodedLiteralNode
        : public LiteralNode
    {
        UnencodedLiteralNode(std::string const &s)
        : LiteralNode(s)
        {
        }

        virtual ~UnencodedLiteralNode() {}

        virtual std::string executeEncapsulated(ExecutionContext *context) const
        {
            return value_;
        }


    };


};  // namespace StringExpr

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

struct StringExpression
    ///
    ///  A representation of a single string which could be represented in the
    ///  bash interpreter.  
    ///
    ///  This class represents the terms of an expression which will 
    ///  generate the proper value given an execution context.
    ///
    ///  The cxxtls::StringExpr::parse() method, defined below, will
    ///  create StringExpression objects.  To executeHelper the string
    ///  expression and get a real string, you use the
    ///  executeHelper() method and pass it an cxxtls::StringExpr::ExecutionContext
    ///  with the environment variables and command line arguments
    ///  properly populated.
    ///
    ///  A StringExpression is a collection of expression terms which are
    ///  executeHelperd in sequence and whose results are concatenated together
    ///  to produce the final string given the ExecutionContext.
    ///
    ///  Here are some examples:
    ///
    ///    abc
    ///    123
    ///    &gt;abc
    ///    "abc"
    ///    abc"1234"def
    ///    &gt; filename
    ///    &gt;&gt; filename
    ///    &lt;stuff
    ///    &lt;'stuff'"morestuff"
    ///    2&gt;stuff
    ///    $varname
    ///    ${varname}
    ///    ${12}
    ///    ${varname%removedFromEnd}
    ///    ${varname#removedFromStart}
    ///    |
    //     || 


{
    typedef StringExpr::ExecutionContext ExecutionContext;

    //==============================================================
    StringExpression() 
    :  useSpaceSeparators_(false)
    ,  protect_(false)
    ,  isSpecial_(false)
    ,  isRedir_(false)
    {

    } 

    std::list<StringExpr::Node*>::iterator begin() { return expression_.begin(); }
    std::list<StringExpr::Node*>::iterator end()   { return expression_.end(); }

    std::list<StringExpr::Node*>::const_iterator begin() const { return expression_.begin(); }
    std::list<StringExpr::Node*>::const_iterator end()   const { return expression_.end(); }


    bool isSpecial() const { return isSpecial_; } ///< return true if this a
                                                  ///  bash statement separator or
                                                  ///  standard file redirection
                                                  
    bool isRedir() const { return isRedir_; }     ///< return true if this a
                                                  ///  bash standard file redirection

    void setSpecial() { isSpecial_ = true; }      ///< mark this expression as 
                                                  ///  statement separator

    void setRedir() { isSpecial_ = true;
                      isRedir_   = true;
                    }
        ///<  mark this string as a bash standard file redirection

    void separateWithSpaces(bool flag=true)
        ///
        ///  Instruct the string expression to either separate the
        ///  individual components with spaces when executing -- or not.
        ///
    {
       useSpaceSeparators_ = flag;   // I am not sure this is ever used
    }

    void protect(bool flag=true)
        ///
        ///  Instruct the string expression to either surround the
        ///  individual components with quotes when executing -- or not.
        ///
    {
       protect_ = flag;
    }


    bool empty() const { return expression_.empty(); }  
        ///<  @return true if the expression is empty in all possible
        ///   execution environments.  By empty, we don't mean, a
        ///   string of length zero.  Rather, we mean that the expression
        ///   no terms, factors, primitives, or whatever.
        ///
        ///   An expression that results in an empty string is not
        ///   "empty", it has one thing in it -- an empty string.

    
    bool isSimple(std::string &value) const
        ///  @return true if the expression contains exactly 1 node and if that
        ///  node is exactly a simple piece of literal text.  If true, return
        ///  the string value of that literal.
    {
        if(1 != expression_.size())
            return false;

        if((*expression_.begin())->isLiteral())
        {
           value = (*expression_.begin())->stringify();

           return true;
        }

        return false;

        
    }


    //==============================================================

    typedef std::pair<size_t, size_t>       CharacterRange;     //  [start,end) 
    typedef std::pair<bool, CharacterRange> RegionDescriptor;   //  protected + range


    std::string executeEncapsulated(ExecutionContext*) const;
        ///<  return a possibly wrapped form of the result.  No globbing or splitting will take place

    void executeHelper(ExecutionContext *context,  StringArray &output, bool glob, bool splitting) const;
        ///<  Given an execution contents containing environment
        ///   variables, command line parameters, and options, 
        ///   produce a string from this expression.

    void executeHelper_1(ExecutionContext *context,  StringArray &output, bool glob, bool splitting, bool encapsulated) const;
        ///<  Given an execution contents containing environment
        ///   variables, command line parameters, and options, 
        ///   produce a string from this expression.

    void executeHelperGlobHelper(ExecutionContext *context, StringArray &output, bool glob, std::string const &rv, std::vector<RegionDescriptor> &regions) const;
        ///<  this function is called by executeHelper() on fully formed strings to expand them using globbing and return the results into the output
        ///   vector.

    void execute(ExecutionContext *context, StringArray &output) const
       ///< this is the top level executor() for string expressions.  Only at the top level does glob = true and splitting = true.  At other
       ///  levels, they are both false.  And glob only equals true if the execution context says it should be true.  For the moment, it is hard coded to
       ///  true.
    {
       executeHelper(context, output, true, true);
    }

    typedef void (*PreGlobFilterFunction)(std::string &value, ExecutionContext *context);  ///<  Data type defining the pre-glob filter function.
                                                                                                 ///   @see preGlobFilter(), below.

    static PreGlobFilterFunction preGlobFilter_;   // see stringexpression.cxx for implementation

    static void registerPreGlobFilterFunction(PreGlobFilterFunction f)
        ///<  This function is used by the StringExpression::execute() methods.  @see function
        ///   preGlobFilter();
    {
        if(f)
        {
           if(preGlobFilter_ == 0)
               preGlobFilter_ = f;
        }
    }

    static void preGlobFilter(std::string &value, ExecutionContext *context)
        ///<  Perform the preGlobFilter operation on the specified string.
        ///
        ///   @param[in,out]  value     The string to be filtered.
        ///   @param[in]      context   The execution context use in performing the
        ///                             filtration.
        ///
        ///  This function's job is to modify the specified spring in such a way
        ///  as to account for the current script language's peculiarities with
        ///  respect to file names.
        ///
        ///  In brash, for example, the tilde character, when appearing at the
        ///  beginning of a string which is to be used as a file name, is interpreted
        ///  as a reference to the user's home directory.
        ///
        ///  This function, preGlobFilter, is invoked just before file name pattern
        ///  expansion is invoked so as to make the conversion.
        ///
        ///  The default preGlobFilter() function does nothing.  To implement a
        ///  language specific filter, you must implement the filter function, then
        ///  call the registerPreGlobFilterFunction() one time during the startup of
        ///  the interpreter you are implementing.  This function must not be changed
        ///  during execution after it is initially set.
    {
        if(preGlobFilter_)
           (*preGlobFilter_)(value,context);
    }




     ~StringExpression();

    //==============================================================
    void adoptNode(StringExpr::Node *n)
    {
       expression_.push_back(n);
    }

    //==============================================================
private:

    //==============================================================
    StringExpr::Expression expression_;

    bool useSpaceSeparators_;
    bool protect_;

         bool isSpecial_;  ///<  this node represents a special string in the
                           ///   bash script languages.  Special nodes fall as follows
                           ///
                           ///      (
                           ///      ((
                           ///      ()
                           ///      )
                           ///      ))
                           ///      [
                           ///      [[
                           ///      ]]
                           ///      ]
                           ///      ;
                           ///      ;;
                           ///      |
                           ///      ||
                           ///      &
                           ///      &&
                           ///      >
                           ///      <

         bool isRedir_;    ///<  contains an unprotected < or >


    //==============================================================
    void operator=(StringExpression const &rhs);  // disable operator=

    //==============================================================
    StringExpression(StringExpression const &);   // disable copy construction

public:
    //==============================================================

    std::string stringify() const;  // debug print helper

    //==============================================================

    template<class Stream>
    friend Stream &operator<< (Stream &s, StringExpression const &me)
    {
         return s << me.stringify();

    }
};  // class StringExpression


}  // namespace cxxtls

#endif
